@@ -1,51 +1,38 @@
 from __future__ import print_function
 
 import os
-import sys
-import numpy as np
-import nibabel as nib
-import tempfile
-import pandas as pd
 import re
+import shutil
+import sys
+import tempfile
 
-try:  # run as a package if installed
-    from pcntoolkit import configs
-except ImportError:
-    pass
+import nibabel as nib
+import numpy as np
+import pandas as pd
 
-    path = os.path.abspath(os.path.dirname(__file__))
-    path = os.path.dirname(path)  # parent directory
-    if path not in sys.path:
-        sys.path.append(path)
-    del path
-    import configs
+from pcntoolkit.util.output import Errors, Messages, Output
 
-CIFTI_MAPPINGS = ('dconn', 'dtseries', 'pconn', 'ptseries', 'dscalar',
-                  'dlabel', 'pscalar', 'pdconn', 'dpconn',
-                  'pconnseries', 'pconnscalar')
+path = os.path.abspath(os.path.dirname(__file__))
+path = os.path.dirname(path)  # parent directory
+if path not in sys.path:
+    sys.path.append(path)
+del path
 
-CIFTI_VOL_ATLAS = 'Atlas_ROIs.2.nii.gz'
+CIFTI_MAPPINGS = (
+    "dconn",
+    "dtseries",
+    "pconn",
+    "ptseries",
+    "dscalar",
+    "dlabel",
+    "pscalar",
+    "pdconn",
+    "dpconn",
+    "pconnseries",
+    "pconnscalar",
+)
 
-PICKLE_PROTOCOL = configs.PICKLE_PROTOCOL
-
-# ------------------------
-# general utility routines
-# ------------------------
-
-
-def predictive_interval(s2_forward,
-                        cov_forward,
-                        multiplicator):
-    """
-    Calculates a predictive interval for the forward model
-    """
-  # calculates a predictive interval
-
-    PI = np.zeros(len(cov_forward))
-    for i, xdot in enumerate(cov_forward):
-        s = np.sqrt(s2_forward[i])
-        PI[i] = multiplicator*s
-    return PI
+CIFTI_VOL_ATLAS = "Atlas_ROIs.2.nii.gz"
 
 
 def create_mask(data_array, mask, verbose=False):
@@ -64,8 +51,7 @@ def create_mask(data_array, mask, verbose=False):
     # create a (volumetric) mask either from an input nifti or the nifti itself
 
     if mask is not None:
-        if verbose:
-            print('Loading ROI mask ...')
+        Output.print(Messages.LOADING_ROI_MASK)
         maskvol = load_nifti(mask, vol=True)
         maskvol = maskvol != 0
     else:
@@ -74,8 +60,7 @@ def create_mask(data_array, mask, verbose=False):
         else:
             dim = data_array.shape[0:3] + (data_array.shape[3],)
 
-        if verbose:
-            print('Generating mask automatically ...')
+        Output.print(Messages.GENERATING_MASK_AUTOMATICALLY)
         if dim[3] == 1:
             maskvol = data_array[:, :, :] != 0
         else:
@@ -128,21 +113,23 @@ def file_type(filename):
                     file_type(filename)
 
     :param filename: name of the file to check
+    :returns: str: The file type ('cifti', 'nifti', 'text', or 'binary')
+    :raises: ValueError if file type is unknown
     """
-    # routine to determine filetype
+    # Map file extensions to types
+    extension_map = {
+        "cifti": (".dtseries.nii", ".dscalar.nii", ".dlabel.nii"),
+        "nifti": (".nii.gz", ".nii", ".img", ".hdr"),
+        "text": (".txt", ".csv", ".tsv", ".asc"),
+        "binary": (".pkl",),
+    }
 
-    if filename.endswith(('.dtseries.nii', '.dscalar.nii', '.dlabel.nii')):
-        ftype = 'cifti'
-    elif filename.endswith(('.nii.gz', '.nii', '.img', '.hdr')):
-        ftype = 'nifti'
-    elif filename.endswith(('.txt', '.csv', '.tsv', '.asc')):
-        ftype = 'text'
-    elif filename.endswith(('.pkl')):
-        ftype = 'binary'
-    else:
-        raise ValueError("I don't know what to do with " + filename)
+    # Check each type's extensions
+    for ftype, extensions in extension_map.items():
+        if any(filename.endswith(ext) for ext in extensions):
+            return ftype
 
-    return ftype
+    raise ValueError(Output.error(Errors.UNKNOWN_FILE_TYPE, filename=filename))
 
 
 def file_extension(filename):
@@ -155,26 +142,7 @@ def file_extension(filename):
 
     :param filename: name of the file to check
     """
-
-    # routine to get the full file extension (e.g. .nii.gz, not just .gz)
-
-    parts = filename.split(os.extsep)
-
-    if parts[-1] == 'gz':
-        if parts[-2] == 'nii' or parts[-2] == 'img' or parts[-2] == 'hdr':
-            ext = parts[-2] + '.' + parts[-1]
-        else:
-            ext = parts[-1]
-    elif parts[-1] == 'nii':
-        if parts[-2] in CIFTI_MAPPINGS:
-            ext = parts[-2] + '.' + parts[-1]
-        else:
-            ext = parts[-1]
-    else:
-        ext = parts[-1]
-
-    ext = '.' + ext
-    return ext
+    return os.path.splitext(filename)[1]
 
 
 def file_stem(filename):
@@ -187,17 +155,15 @@ def file_stem(filename):
 
     :param filename: name of the file to check
     """
-    idx = filename.find(file_extension(filename))
-    stm = filename[0:idx]
+    return os.path.splitext(os.path.basename(filename))[0]
 
-    return stm
 
 # --------------
 # nifti routines
 # --------------
 
 
-def load_nifti(datafile, mask=None, vol=False, verbose=False):
+def load_nifti(datafile, mask=None, vol=False):
     """
     Load a nifti file into a numpy array
 
@@ -211,10 +177,8 @@ def load_nifti(datafile, mask=None, vol=False, verbose=False):
     :param verbose: verbose output
     """
 
-    if verbose:
-        print('Loading nifti: ' + datafile + ' ...')
     img = nib.load(datafile)
-    dat = img.get_data()
+    dat = img.get_fdata()
 
     if mask is not None:
         mask = load_nifti(mask, vol=True)
@@ -226,7 +190,7 @@ def load_nifti(datafile, mask=None, vol=False, verbose=False):
 
 
 def save_nifti(data, filename, examplenii, mask, dtype=None):
-    '''
+    """
     Write output to nifti
 
     Basic usage::
@@ -238,7 +202,7 @@ def save_nifti(data, filename, examplenii, mask, dtype=None):
     :param examplenii: nifti to copy the geometry and data type from
     :mask: nifti image containing a mask for the image
     :param dtype: data type for the output image (if different from the image)
-    '''
+    """
 
     # load mask
     if isinstance(mask, str):
@@ -258,7 +222,7 @@ def save_nifti(data, filename, examplenii, mask, dtype=None):
     # write data
     array_data = np.zeros((np.prod(dim), nvol))
     array_data[mask.flatten(), :] = data
-    array_data = np.reshape(array_data, dim+(nvol,))
+    array_data = np.reshape(array_data, dim + (nvol,))
     hdr = ex_img.header
     if dtype is not None:
         hdr.set_data_dtype(dtype)
@@ -267,6 +231,7 @@ def save_nifti(data, filename, examplenii, mask, dtype=None):
 
     nib.save(array_img, filename)
 
+
 # --------------
 # cifti routines
 # --------------
@@ -274,7 +239,7 @@ def save_nifti(data, filename, examplenii, mask, dtype=None):
 
 def load_cifti(filename, vol=False, mask=None, rmtmp=True):
     """
-    Load a cifti file into a numpy array 
+    Load a cifti file into a numpy array
 
     Basic usage::
 
@@ -288,17 +253,14 @@ def load_cifti(filename, vol=False, mask=None, rmtmp=True):
     # parse the name
     dnam, fnam = os.path.split(filename)
     fpref = file_stem(fnam)
-    outstem = os.path.join(tempfile.gettempdir(),
-                           str(os.getpid()) + "-" + fpref)
+    outstem = os.path.join(tempfile.gettempdir(), str(os.getpid()) + "-" + fpref)
 
     # extract surface data from the cifti file
-    print("Extracting cifti surface data to ", outstem, '-*.func.gii', sep="")
-    giinamel = outstem + '-left.func.gii'
-    giinamer = outstem + '-right.func.gii'
-    os.system('wb_command -cifti-separate ' + filename +
-              ' COLUMN -metric CORTEX_LEFT ' + giinamel)
-    os.system('wb_command -cifti-separate ' + filename +
-              ' COLUMN -metric CORTEX_RIGHT ' + giinamer)
+    Output.print(Messages.EXTRACTING_CIFTI_SURFACE_DATA, outstem=outstem)
+    giinamel = outstem + "-left.func.gii"
+    giinamer = outstem + "-right.func.gii"
+    os.system("wb_command -cifti-separate " + filename + " COLUMN -metric CORTEX_LEFT " + giinamel)
+    os.system("wb_command -cifti-separate " + filename + " COLUMN -metric CORTEX_RIGHT " + giinamer)
 
     # load the surface data
     giil = nib.load(giinamel)
@@ -306,8 +268,7 @@ def load_cifti(filename, vol=False, mask=None, rmtmp=True):
     Nimg = len(giil.darrays)
     Nvert = len(giil.darrays[0].data)
     if Nimg == 1:
-        out = np.concatenate((giil.darrays[0].data, giir.darrays[0].data),
-                             axis=0)
+        out = np.concatenate((giil.darrays[0].data, giir.darrays[0].data), axis=0)
     else:
         Gl = np.zeros((Nvert, Nimg))
         Gr = np.zeros((Nvert, Nimg))
@@ -321,10 +282,9 @@ def load_cifti(filename, vol=False, mask=None, rmtmp=True):
         os.remove(giinamer)
 
     if vol:
-        niiname = outstem + '-vol.nii'
-        print("Extracting cifti volume data to ", niiname, sep="")
-        os.system('wb_command -cifti-separate ' + filename +
-                  ' COLUMN -volume-all ' + niiname)
+        niiname = outstem + "-vol.nii"
+        Output.print(Messages.EXTRACTING_CIFTI_VOLUME_DATA, niiname=niiname)
+        os.system("wb_command -cifti-separate " + filename + " COLUMN -volume-all " + niiname)
         vol = load_nifti(niiname, vol=True)
         volmask = create_mask(vol)
         out = np.concatenate((out, vol2vec(vol, volmask)), axis=0)
@@ -351,13 +311,11 @@ def save_cifti(data, filename, example, mask=None, vol=True, volatlas=None):
     """
 
     # do some sanity checks
-    if data.dtype == 'float32' or \
-       data.dtype == 'float' or \
-       data.dtype == 'float64':
-        data = data.astype('float32')  # force 32 bit output
-        dtype = 'NIFTI_TYPE_FLOAT32'
+    if data.dtype == "float32" or data.dtype == "float" or data.dtype == "float64":
+        data = data.astype("float32")  # force 32 bit output
+        dtype = "NIFTI_TYPE_FLOAT32"
     else:
-        raise ValueError('Only float data types currently handled')
+        raise ValueError(Output.error(Errors.NO_FLOAT_DATA_TYPE, data_type=data.dtype))
 
     if len(data.shape) == 1:
         Nimg = 1
@@ -371,23 +329,19 @@ def save_cifti(data, filename, example, mask=None, vol=True, volatlas=None):
 
     # Split the template
     estem = os.path.join(tempfile.gettempdir(), str(os.getpid()) + "-" + fstem)
-    giiexnamel = estem + '-left.func.gii'
-    giiexnamer = estem + '-right.func.gii'
-    os.system('wb_command -cifti-separate ' + example +
-              ' COLUMN -metric CORTEX_LEFT ' + giiexnamel)
-    os.system('wb_command -cifti-separate ' + example +
-              ' COLUMN -metric CORTEX_RIGHT ' + giiexnamer)
+    giiexnamel = estem + "-left.func.gii"
+    giiexnamer = estem + "-right.func.gii"
+    os.system("wb_command -cifti-separate " + example + " COLUMN -metric CORTEX_LEFT " + giiexnamel)
+    os.system("wb_command -cifti-separate " + example + " COLUMN -metric CORTEX_RIGHT " + giiexnamer)
 
     # write left hemisphere
     giiexl = nib.load(giiexnamel)
     Nvertl = len(giiexl.darrays[0].data)
     garraysl = []
     for i in range(0, Nimg):
-        garraysl.append(
-            nib.gifti.gifti.GiftiDataArray(data=data[0:Nvertl, i],
-                                           datatype=dtype))
+        garraysl.append(nib.gifti.gifti.GiftiDataArray(data=data[0:Nvertl, i], datatype=dtype))
     giil = nib.gifti.gifti.GiftiImage(darrays=garraysl)
-    fnamel = fstem + '-left.func.gii'
+    fnamel = fstem + "-left.func.gii"
     nib.save(giil, fnamel)
 
     # write right hemisphere
@@ -395,41 +349,48 @@ def save_cifti(data, filename, example, mask=None, vol=True, volatlas=None):
     Nvertr = len(giiexr.darrays[0].data)
     garraysr = []
     for i in range(0, Nimg):
-        garraysr.append(
-            nib.gifti.gifti.GiftiDataArray(data=data[Nvertl:Nvertl+Nvertr, i],
-                                           datatype=dtype))
+        garraysr.append(nib.gifti.gifti.GiftiDataArray(data=data[Nvertl : Nvertl + Nvertr, i], datatype=dtype))
     giir = nib.gifti.gifti.GiftiImage(darrays=garraysr)
-    fnamer = fstem + '-right.func.gii'
+    fnamer = fstem + "-right.func.gii"
     nib.save(giir, fnamer)
 
     tmpfiles = [fnamer, fnamel, giiexnamel, giiexnamer]
 
     # process volumetric data
     if vol:
-        niiexname = estem + '-vol.nii'
-        os.system('wb_command -cifti-separate ' + example +
-                  ' COLUMN -volume-all ' + niiexname)
+        niiexname = estem + "-vol.nii"
+        os.system("wb_command -cifti-separate " + example + " COLUMN -volume-all " + niiexname)
         niivol = load_nifti(niiexname, vol=True)
         if mask is None:
             mask = create_mask(niivol)
 
         if volatlas is None:
             volatlas = CIFTI_VOL_ATLAS
-        fnamev = fstem + '-vol.nii'
+        fnamev = fstem + "-vol.nii"
 
-        save_nifti(data[Nvertr+Nvertl:, :], fnamev, niiexname, mask)
+        save_nifti(data[Nvertr + Nvertl :, :], fnamev, niiexname, mask)
         tmpfiles.extend([fnamev, niiexname])
 
     # write cifti
-    fname = fstem + '.dtseries.nii'
-    os.system('wb_command -cifti-create-dense-timeseries ' + fname +
-              ' -volume ' + fnamev + ' ' + volatlas +
-              ' -left-metric ' + fnamel + ' -right-metric ' + fnamer)
+    fname = fstem + ".dtseries.nii"
+    os.system(
+        "wb_command -cifti-create-dense-timeseries "
+        + fname
+        + " -volume "
+        + fnamev
+        + " "
+        + volatlas
+        + " -left-metric "
+        + fnamel
+        + " -right-metric "
+        + fnamer
+    )
 
     # clean up
     for f in tmpfiles:
         os.remove(f)
 
+
 # --------------
 # ascii routines
 # --------------
@@ -447,9 +408,7 @@ def load_pd(filename):
     """
 
     # based on pandas
-    x = pd.read_csv(filename,
-                    sep=' ',
-                    header=None)
+    x = pd.read_csv(filename, sep=" ", header=None)
     return x
 
 
@@ -465,11 +424,7 @@ def save_pd(data, filename):
     :param filename: where to store it
     """
     # based on pandas
-    data.to_csv(filename,
-                index=None,
-                header=None,
-                sep=' ',
-                na_rep='NaN')
+    data.to_csv(filename, index=None, header=None, sep=" ", na_rep="NaN")
 
 
 def load_ascii(filename):
@@ -502,6 +457,7 @@ def save_ascii(data, filename):
     # based on pandas
     np.savetxt(filename, data)
 
+
 # ----------------
 # generic routines
 # ----------------
@@ -523,20 +479,22 @@ def save(data, filename, example=None, mask=None, text=False, dtype=None):
     :param dtype: data type for the output image (if different from the image)
     """
 
-    if file_type(filename) == 'cifti':
+    if file_type(filename) == "cifti":
         save_cifti(data.T, filename, example, vol=True)
-    elif file_type(filename) == 'nifti':
+    elif file_type(filename) == "nifti":
         save_nifti(data.T, filename, example, mask, dtype=dtype)
-    elif text or file_type(filename) == 'text':
+    elif text or file_type(filename) == "text":
         save_ascii(data, filename)
-    elif file_type(filename) == 'binary':
+    elif file_type(filename) == "binary":
         data = pd.DataFrame(data)
         data.to_pickle(filename, protocol=PICKLE_PROTOCOL)
 
 
 def load(filename, mask=None, text=False, vol=True):
     """
-    Load a numpy array from a file
+    Load array from a file.
+
+    A beautiful waterfall of errors.
 
     Basic usage::
 
@@ -547,62 +505,60 @@ def load(filename, mask=None, text=False, vol=True):
     :param text: whether to write out a text file
     :param vol: whether to load the image as a volume
     """
-
-    if file_type(filename) == 'cifti':
-        x = load_cifti(filename, vol=vol)
-    elif file_type(filename) == 'nifti':
-        x = load_nifti(filename, mask, vol=vol)
-    elif text or file_type(filename) == 'text':
-        x = load_ascii(filename)
-    elif file_type(filename) == 'binary':
-        x = pd.read_pickle(filename)
-        x = x.to_numpy()
-    return x
-
-# -------------------
-# sorting routines for batched in normative parallel
-# -------------------
-
-
-def tryint(s):
-    """
-    Try to convert a string to an integer
-
-    Basic usage::
-
-                    tryint(s)
-
-    :param s: string to convert
-    """
-
+    if not os.path.exists(filename):
+        raise FileNotFoundError(f"File not found: {filename} ")
     try:
-        return int(s)
-    except ValueError:
-        return s
+        x = pd.read_pickle(filename)
+        return x.to_numpy()
+    except Exception:
+        try:
+            x = np.loadtxt(filename)
+            return x
+        except Exception:
+            try:
+                x = load_ascii(filename)
+                return x
+            except Exception:
+                try:
+                    x = load_nifti(filename, mask, vol=vol)
+                    return x
+                except Exception:
+                    try:
+                        x = load_cifti(filename, vol=vol)
+                        return x
+                    except Exception:
+                        raise ValueError(Output.error(Errors.UNKNOWN_FILE_TYPE, filename=filename))
 
 
-def alphanum_key(s):
+def create_incremental_backup(filepath):
     """
-    Turn a string into a list of numbers
+    Create an incremental backup of a file using the `.bak{n}` naming scheme.
 
-    Basic usage::
-
-                    alphanum_key(s) 
-
-    :param s: string to convert
+    If the file does not exist, an empty file is created at the specified path.
+    A backup is then created in the same directory with the following format:
+        original_name.bak{n}.original_extension
+    where {n} is incremented based on existing backups.
     """
-    return [tryint(c) for c in re.split('([0-9]+)', s)]
+    directory, filename = os.path.split(filepath)
+    name, ext = os.path.splitext(filename)
 
+    if not os.path.exists(filepath):
+        # Create an empty file
+        with open(filepath, "w") as f:
+            pass
+    else:
+        print("File exists:")
 
-def sort_nicely(l):
-    """
-    Sort a list of strings in a natural way
+    # Pattern: name.bakN.ext
+    regex = re.compile(rf"^{re.escape(name)}\.bak(\d+){re.escape(ext)}$")
 
-    Basic usage::
+    existing = [f for f in os.listdir(directory) if regex.match(f)]
 
-        sort_nicely(l)  
+    numbers = [int(regex.match(f).group(1)) for f in existing] if existing else []
 
-    :param l: list of strings to sort
-    """
+    next_n = max(numbers, default=0) + 1
+    backup_name = f"{name}.bak{next_n}{ext}"
+    backup_path = os.path.join(directory, backup_name)
 
-    return sorted(l, key=alphanum_key)
+    shutil.copy2(filepath, backup_path)
+    return backup_path
